#include <stdint.h>

#include <stdio.h>

#include <cstring>
#include <string>
#include <map>
#include <vector>
#include <cmath>

#include <list>

#define OPCODE_DEBUG 1
#define VERIFY_RAM   1

static bool OPCODE_DEBUG_ENABLED = false;
static bool TIMER_ENABLED        = true;
static bool DEBUG_FORCED = false;

# define DISASM_NOTSTANDALONE
# include "microblaze-dis.cc"

#include <functional>

/* Declare integer datatypes for each number of bits */
typedef uint_least8_t  u8;  typedef int_least8_t  s8;
typedef uint_least16_t u16; typedef int_least16_t s16;
typedef uint_least32_t u32; typedef int_least32_t s32;
typedef uint_least64_t u64; typedef int_least64_t s64;

/* The following data structures provide templated access
 * of different data sizes within a 32-bit integer.
 */
template<unsigned size, typename T=void>
struct MemDataRef
{
    union { u32 value; T alt[4 / sizeof(T)]; } d;
    T& reffun(unsigned ind) { return d.alt[ind >> (sizeof(T)==2)]; }
};
template<> struct MemDataRef<1,void>: public MemDataRef<0, u8>  { };
template<> struct MemDataRef<2,void>: public MemDataRef<0, u16> { };
template<> struct MemDataRef<4,void>: public MemDataRef<0, u32> { };

/* Declare facilities for detecting and dealing with different byteorder */
#include <endian.h>
#define LITTLE_ENDIAN_HOST  (__BYTE_ORDER == __LITTLE_ENDIAN)
#define LITTLE_ENDIAN_SLAVE 0 /* 0=big endian */
#define NOfprintf(...) /**/
#define CROSS_ENDIAN (LITTLE_ENDIAN_HOST != LITTLE_ENDIAN_SLAVE)

static unsigned SwapBytes(unsigned value, unsigned size)
{
    if(size >= 2) value = ((value & 0xFF00FF00u) >> 8)
                        | ((value & 0x00FF00FFu) << 8);
    if(size >= 4) value = (value >> 16) | (value << 16);
    return value;
}

struct Syscall
{
    unsigned no;
    std::string name;
    struct param { std::string name; enum { m_int, m_long, m_charp, m_charpp, m_ptr } type; };
    std::vector< param > params;
};
std::map<unsigned, Syscall> syscalls;

struct
{
    u32 regs[32], pid;
    u32 return_pc;
} active_syscall;


#include <termio.h>
#include <sys/ioctl.h>
#include <fcntl.h>
static class Console
{
    struct termio back;
    int pending;
public:
    Console() : pending(0)
    {
        ioctl(0, TCGETA, &back);
        struct termio term = back;
        // Disable linebuffer and echoing
        term.c_lflag &= ~(ICANON|ECHO);
        term.c_cc[VMIN] = 0; // 0=no block, 1=do block
        if(ioctl(0, TCSETA, &term) < 0)
            fcntl(0, F_SETFL, fcntl(0, F_GETFL) | O_NONBLOCK);
    }
    ~Console()
    {
        if(ioctl(0, TCSETA, &back) < 0)
            fcntl(0, F_SETFL, fcntl(0, F_GETFL) &~ O_NONBLOCK);
    }
    bool Hit()
    {
        if(pending) return true;
        char c;
        int r = read(0, &c, 1);
        if(r > 0) { pending = c; return true; }
        return false;
    }
    unsigned Getc() { int r = pending;
      if(r=='')
      {
        DEBUG_FORCED = !DEBUG_FORCED;
        fprintf(stdout, "DEBUG SET TO %d\n", DEBUG_FORCED);
        fprintf(stderr, "DEBUG SET TO %d\n", DEBUG_FORCED);
      }
      pending = 0; return r; }
    void Putc(unsigned c) { putchar(c); fflush(stdout); }
} Console;

/* Declare all the devices we emulate. */

// RAM device is very simple.
template<unsigned size_bytes>
class SDRAM
{
    u32 buffer[size_bytes / 4];
public:
    u32 Read(u32 index) const        { return buffer[index]; }
    void Write(u32 index, u32 value) { buffer[index] = value; }
};

// Xilinx OPB Interrupt Controller
class XilinxINTC
{
    // List of registers
    enum { ISR/* List of active IRQs */,
           IPR,
           IER/* List of IRQs that can result in an interrupt */,
           IAR, SIE, CIE, IVR, MER };
    // List of bits
    enum { ME=1, HIE=2 };
    u32 regs[8];
public:
    XilinxINTC(): regs() { }
    u32 Read(u32 index) const
    {
        if(index == IPR) return regs[ISR] & regs[IER];
        if(index == IVR) // Ordinal number of highest enabled interrupt
        {
            for(unsigned i=0; i<32; ++i) if(Read(IPR) & (1<<i)) return i;
            return ~0u;
        }
        return regs[index];
    }
    void Write(u32 index, u32 value)
    {
        switch(index)
        {
            case IAR: regs[ISR] &= ~value; break; // Acknowledge
            case SIE:
                fprintf(stderr, "INTC: Enable IRQ mask %02X\n", value);
                regs[IER] |=  value;
                break; // Atomic set ie.
            case CIE:
                fprintf(stderr, "INTC: Disable IRQ mask %02X\n", value);
                regs[IER] &= ~value;
                break; // Atomic clear ie.
            case ISR: if(!(regs[MER] & HIE))    // If software interrupts are enabled,
            default:       regs[index] = value; // Update the register as indicated.
        }
        if(index == MER)
            fprintf(stderr, "INTC: MER set to ME=%d, HIE=%d\n", value&ME, value&HIE);
        NOfprintf(stderr, "INTC WRITE[%u,0x%X]: Enabled are %X, set are %X\n",
            index,value,regs[IER],regs[ISR]);
    }
    bool CheckIRQ() { return (regs[MER] & ME) && Read(IPR); }
    void TriggerIRQ(unsigned irq)
    {
        if(regs[MER] & HIE) // If hardware interrupts are enabled,
        {
            if(!(regs[ISR] & (1 << irq)))
            {
                fprintf(stderr, "INTC: setting IRQ level %u (mask %02X)\n", irq, 1 << irq);
            }
            regs[ISR] |= 1 << irq; // set the IRQ level.
        }
    }
};

// Xilinx interrupt timer
template<unsigned N>
class XilinxTimer
{
    // List of registers
    enum { TCSR=0, TLR=1, TCR=2 };
    // List of control bits
    enum { UDT=2, ARHT=16, LOAD=32, ENIT=64, ENT=128, TINT=256, ENALL=1024 };
    struct impl
    {
        u32 regs[4];
        void Write(u32 index, u32 value)
        {
            if(index == TCSR)
                { // Setting the TINT bit actually clears it,
                  // otherwise it is preserved.
                  if(value & TINT) value &= ~TINT;
                  else value |= (regs[index] & TINT); }
            regs[index] = value;
            NOfprintf(stderr, "TIMER WRITE[%u,0x%X]: Ctrl=%X,Count=%u,%s,%s,%s\n",
                index,value, regs[TCSR], regs[TCR],
                    (regs[TCSR]&UDT)?"down":"up",
                    (regs[TCSR]&ENT)?"enabled":"disabled",
                    (regs[TCSR]&LOAD)?"load":
                     (regs[TCSR]&ARHT)?"autoload":"");
        }
        bool CheckIRQ() const { return (regs[TCSR] & (TINT|ENIT)) == (TINT|ENIT); }
        void Tick()
        {
            if(!(regs[TCSR] & ENT)) return; // Timer does nothing unless enabled
            if(regs[TCSR] & LOAD) { regs[TCR] = regs[TLR]; return; }
            u32 previous = regs[TCSR & UDT] ? regs[TCR]-- : regs[TCR]++;
            if( (previous ^ regs[TCR]) == ~u32(0)) // overflow in either direction
            {
                //fprintf(stderr, "TIMER overflow (%u,%u)\n", previous,regs[TCR]);
                regs[TCSR] |= TINT;
                if(regs[TCSR] & ARHT) regs[TCR] = regs[TLR]; // Restart timer
                else                  regs[TCSR] &= ~ENT;    // Stop timer
            }
        }
    } timers[N];
public:
    u32 Read(u32 which, u32 index) { return timers[which].regs[index]; }
    void Write(u32 which, u32 index, u32 value)
    {
        timers[which].Write(index, value);
        // Setting the ENALL bit on any of the timers sets ENT on all of them.
        if(value & ENALL) for(auto& t: timers) t.regs[TCSR] |= ENT;
    }
    void Tick() { for(auto& t: timers) t.Tick(); }
    bool CheckIRQ() const
    {
        // Level-triggered.
        for(const auto& t: timers) if(t.CheckIRQ()) return true;
        return false;
    }
};

// Xilinx UART lite, lite version
class XilinxUARTlite
{
    bool ints_enabled, tx_emptied_int_pending, overrun;
    struct { u8 len, pos, fifo[8]; } in;
    // List of supported status and control bits
    enum { RXVALID=1, RXFULL=2, TXEMPTY=4, IE=16, OVERRUN=32, RST_RX=2 };
public:
    XilinxUARTlite(): ints_enabled(0), overrun(0), in{0,0,{0}} { }
    u32 Read(u32 index)
    {
        u32 result = TXEMPTY;
        if(ints_enabled) result |= IE;
        if(in.len > 0)   result |= RXVALID;
        if(in.len == 8)  result |= RXFULL;
        if(overrun)     {result |= OVERRUN; overrun=false;}
        if(index<2 && in.len) result = in.fifo[ (in.pos + 8 - in.len--) % 8 ];
        return result;
    }
    void Write(u32 index, u32 value)
    {
        if(index<3) { Console.Putc(value); tx_emptied_int_pending = true; }
        else { ints_enabled = value&IE; if(value&RST_RX) in.pos=in.len=0; }
    }
    bool CheckIRQ() // TX interrupt is an edge; RX interrupt is a level.
    {
        if(ints_enabled && (in.len || tx_emptied_int_pending))
            { tx_emptied_int_pending=false; return true; }
        return false;
    }
    void Input()
    {
        static unsigned counter = 0;
        if(!counter) counter = 100; else { --counter; return; }
        if(!Console.Hit()) return;
        if(in.len >= 8) { overrun = true; return; }
        in.fifo[ in.pos++ % 8 ] = Console.Getc();
        ++in.len;
    }
};

enum intent { intent_load=0, intent_store=1, intent_execute=2 };

class MicroBlazeCPU // Micro-Blaze CPU
{
    enum {
        v_reset  = 0x00,
        v_usr_ex = 0x08,
        v_intr   = 0x10,
        v_brk    = 0x18,
        v_hw_ex  = 0x20
    };

    union reg // MicroBlaze register file
    {
        struct {  // As 32-bit integers
            u32 MSR; // Machine Status Register, i.e. flags
            u32 ESR; // Exception Status Register, i.e. exception flags
            // 32 general purpose registers.  [0] is always zero and cannot be modified.
            u32 r[32];
            // Special purpose registers.
            // See SPR() below for the list of meanings.
            u32 PC, EAR, BTR, SLR, SHR, ZPR, TLBX, PVRs[16];
            union
            {
                struct { u32 LO, HI; } w;
                struct { u32 G:1, M:1, I:1, W:1, ZSEL:4, WR:1,EX:1, RPN:22;
                         // Bits M, I, W are ignored. RPN=Real Page NUmber
                         u32 filler:4, U0:1, E:1, V:1, SIZE:3, TAG:22;
                         // E=endianess. Page size is 4^SIZE kibibytes.
                       } b;
            } UTLB[64];
            u8 TID[64], PID;
        };
        struct { // Or broken down to individual bits:
            // MSR bits:
            u32 bit31:1;
            u32 IE:1;  // bit 30 (Interrupt Enable)
            u32 C:1;   // bit 29 (carry)
            u32 BIP:1; // bit 28 (Break In Progress)
            u32 FSL:1, ICE:1;
            u32 DZO:1; // bit 25 (Division By Zero)
            u32 DCE:1;
            u32 EE:1;  // bit 23 (Exceptions Enabled)
            u32 EIP:1; // bit 22 (Exception In Progress)
            u32 PVR:1; // bit 21 (PVR register set exists)
            u32 UM:1;  // bit 20 (1=Privileged instructions are prohibited, 0=Allowed)
            u32 UMS:1; // bit 19 (User Mode Save)
            u32 VM:1;  // bit 18 (1=MMU address translation and access protection enabled)
            u32 VMS:1; // bit 17 (Virtual Protected Mode Save)
            u32 res:16;
            u32 CC:1;  // bit 0  (carry copy)
            // ESR bits:
            u32 EC: 5;  // Exception cause
             u32 Rx: 5; // Exception specific status: register
             u32 S:  1; // Exception specific status: store
             u32 DIZ:1; // Exception specific status: zone protection
            u32 DS: 1;  // Delay Slot Exception
            u32 filler: 19;
        };

        void PushUVM() { VMS=VM; UMS=UM; VM=0; UM=0; }
        void PopUVM() { VM=VMS; UM=UMS; }
        void SPR(unsigned index, bool write, u32& param)
        {
            #define r(reg, writeok) \
                if(!writeok && write) goto def; \
                if(write) reg = param; else param = reg; \
                return
            if(write)
                { NOfprintf(stderr, "SPR(%04X): write %08X\n", index, param); }

            switch(index)
            {
                case 0x0000: r(PC, 0);     // PC, program counter
                case 0x0001: r(MSR, 1);    // MSR, machine status register (i.e. flags)
                case 0x0003: r(EAR, 0);    // EAR, exception address register
                case 0x0005: r(ESR, 0);    // ESR, exception status register
                case 0x000B: r(BTR, 0);    // BTR, branch target register
                case 0x0800: r(SLR, 1);    // SLR, stack low register
                case 0x0802: r(SHR, 1);    // SHR, stack high register
                case 0x1000: r(PID, 1);    // PID, process identifier register
                case 0x1001: r(ZPR, 1);    // ZPR, zone protection register
                case 0x1002: r(TLBX, 1);   // TLBX, translation look-aside buffer index register
                case 0x1003:               // TLBLO, translation look-aside buffer low register
                    if(write)
                        fprintf(stderr, "TLBLO[%2u] set to %08X\n",
                            (unsigned)(TLBX&63), (unsigned)param);
                    r(UTLB[TLBX&63].w.LO, 1);
                case 0x1004:               // TLBHI, translation look-aside buffer high register
                    if(write)
                    {
                        TID[TLBX&63] = PID;
                        UTLB[TLBX&63].w.HI  = param;
                        if(write)
                            fprintf(stderr, "TLBHI[%2u] set to %08X; TID = %02X\n",
                                (unsigned)(TLBX&63), (unsigned)param, (unsigned)TID[TLBX&63]);
                    }
                    else
                    {
                        if(!UM) PID = TID[TLBX&63];
                        param = UTLB[TLBX&63].w.HI;
                    }
                    return;
                case 0x1005:             // TLBSX, translation look-aside buffer search index register
                    if(write)            // Do a TLB search and report the matching TLB index in TLBX.
                      { TLBX = 0x80000000u; // If nothing found, this value remains indicating the miss.
                        fprintf(stderr, "TLBSX: Searching for %08X\n", param);
                        TLBfind<intent_load,false>(param&0xFFFFFC00,
                          [&TLBX](unsigned tlbno) -> unsigned { TLBX = tlbno; return 0; } );
                        fprintf(stderr, "TLBSX result: status=%08X\n", TLBX);
                        return; }

                default: def:
                    if(index < 0x2000 || write) index = 15; // Fallback to a safe "undefined" slot
                    r(PVRs[index&15], 1); //  processor version registers
            }
            #undef wp
        }

        /* Virtual memory management unit (MMU) functions */
        template<intent rw, bool debug, typename T>
        unsigned TLBfind(unsigned virt_addr, const T& func) const
        {
            // Cache the last used index for each type of access
            static unsigned last_matched_index[3] = {0,0,0};

            if(debug)
            fprintf(stderr, "TLB looking up:   PID=%02X,VIR=%08X (intent=%s)\n",
                PID,
                virt_addr,
                rw==intent_load ? "read"
              : rw==intent_store ? "write"
              : "execute");

            bool UserMode = UM;
            unsigned exception_no = 16+2+4; // Exception for "no tlb"
            for(unsigned vv=0; vv<64; ++vv)
            {
                unsigned v = (last_matched_index[rw] + vv) % 64;

                if(debug)
                fprintf(stderr,
                    "TLB %2u: %s %s %s %s TID=%02X,TAG=%08X,MASK=%08X,PHYS=%08X,ZSEL=%u(%u=%15s) ",
                        v,
                        UTLB[v].b.V? "V":"-",
                        UTLB[v].b.WR?"WR":"--",
                        UTLB[v].b.EX?"EX":"--",
                        UTLB[v].b.G ?"G":"-",
                        TID[v],
                        UTLB[v].b.TAG<<10,
                        ~( (1 << (10 + UTLB[v].b.SIZE*2)) - 1 ),
                        UTLB[v].b.RPN<<10,
                        UTLB[v].b.ZSEL,
                        (ZPR >> (30 - UTLB[v].b.ZSEL*2)) & 3,
                        "Deny Usermode\0 "
                        "Deny Nobody\0   "
                        "Grant Kernel\0  "
                        "Grant Everyone\0"
                        + 15 * ((ZPR >> (30 - UTLB[v].b.ZSEL*2)) & 3));

                if(!UTLB[v].b.V)
                {
                    if(debug) fprintf(stderr, ": invalid\n");
                    continue; // skip invalid tlbs
                }
                if(TID[v] != 0 && PID != TID[v])
                {
                    if(debug)fprintf(stderr, ": mismatch: TID\n");
                    continue;
                }
                unsigned page_bits = 10 + UTLB[v].b.SIZE * 2;
                if( u32(virt_addr ^ (UTLB[v].b.TAG<<10)) >> page_bits )
                {
                    if(debug)fprintf(stderr, ": mismatch: TAG\n");
                    continue;
                }
                unsigned zone_type = (ZPR >> (30 - UTLB[v].b.ZSEL*2)) & 3;
                exception_no &= ~2; // Remove the bit indicating "no tlb found" exception
                if(zone_type == 0 && UserMode)              { if(debug)fprintf(stderr, ": no V\n"); continue; }
                exception_no &= ~4; // Remove the bit indicating "zone protection" exception
                if(zone_type < (UserMode?3:2))
                {
                    if(rw==intent_store   && !UTLB[v].b.WR) { if(debug)fprintf(stderr, ": no WR\n"); continue; }
                    if(rw==intent_execute && !UTLB[v].b.EX) { if(debug)fprintf(stderr, ": no EX\n"); continue; }
                }
                if(rw==intent_execute     && UTLB[v].b.G)   { if(debug)fprintf(stderr, ": is Guarded\n"); continue; }
                last_matched_index[rw] = v;

                if(debug)fprintf(stderr, ": OK\n");
                return func(v);
            }
            return exception_no;
        }
    } regs;

public:
    std::pair<  std::function<u32(u32)>,
                std::function<void(u32,u32)>
             > IOmap[0x10000];

/*
MEMORY ACCESS:

                               Bytes in memory.
                               A0 = least significant byte of data item [addr+0]
                               A1 = least significant byte of data item [addr+1]
                               A2 = least significant byte of data item [addr+2]
                               A3 = least significant byte of data item [addr+3]
                               B0 = second byte of data item [addr+0]
                               B2 = second byte of data item [addr+2]
                               C0 = third byte of data item [addr+0]
                               D0 = most significant byte of data item [addr+0]

SIZE=1,          GUEST=BIG   : A0 A1 A2 A3
SIZE=1,          GUEST=LITTLE: A0 A1 A2 A3
SIZE=1, REVERSE, GUEST=BIG   : A3 A2 A1 A0
SIZE=1, REVERSE, GUEST=LITTLE: A3 A2 A1 A0

SIZE=2,          GUEST=BIG   : B0 A0 B2 A2
SIZE=2,          GUEST=LITTLE: A0 B0 A2 B2
SIZE=2, REVERSE, GUEST=BIG   : A2 B2 A0 B0
SIZE=2, REVERSE, GUEST=LITTLE: B2 A2 B0 A0

SIZE=4,          GUEST=LITTLE: A0 B0 C0 D0
SIZE=4, REVERSE, GUEST=BIG   : A0 B0 C0 D0
SIZE=4,          GUEST=BIG   : D0 C0 B0 A0
SIZE=4, REVERSE, GUEST=LITTLE: D0 C0 B0 A0

*/
    /* All memory access is routed through this template function. */
    /* It handles reads and writes of different sizes,
     * with or without virtual memory mapping,
     * with or without byte order swaps (reverse).
     */
    template<intent rw=intent_load, unsigned size=4>
    unsigned MemAccess(u32 virt_addr, u32& value, bool reverse=false)
    {
        u32 phys_addr = virt_addr;
        bool    e_bit = false;

        if(virt_addr >= 0xC0000000u && virt_addr < 0xC2000000u) // Kernel memory
        {
            // A little anointing hack. It's a bit easier
            // to debug when not constantly interrupted
            // by paging faults.
            phys_addr -= (0xC0000000u - 0x10000000u);
        }else
        /*if(virt_addr >= 0x10000000u && virt_addr < 0x102ECD68u && !regs.UM) // Kernel memory
        {
            // The same, but for UM
        }else*/
        if(regs.VM) // Virtual memory mode? Do TLB lookup.
        {
            unsigned exception = regs.TLBfind<rw,false>(virt_addr,
                [&](unsigned tlbno) -> unsigned
                {
                    // Map the address using the matching TLB.
                    unsigned page_mask = (1024 << (2 * regs.UTLB[tlbno].b.SIZE)) - 1;
                    phys_addr = (regs.UTLB[tlbno].b.RPN << 10)
                              + (virt_addr & page_mask);
                    e_bit = regs.UTLB[tlbno].b.E;
                    return 0;
                });
            if(exception)
            {
                // regs.TLBfind<rw,true>(virt_addr, [](unsigned){ return 0; });
                // Produce an exception if the mapping failed (LSB indicates code/data).
                return exception | (rw==intent_execute);
            }
        }

#if C_UNALIGNED_EXCEPTIONS || 1
        if(virt_addr % size != 0) return 1 | (size==4 ? 4 : 0); // Check alignment
#endif
        try
        {
            // An "E" bit in a TLB reverses the endianess for that page (except in opcode lookups).
            if(rw != intent_execute) reverse ^= e_bit;
            // This macro byteswaps if needed due to host/guest endian differences or the reverse bit.
            auto S = [=](u32 v) -> u32
                { return size>1 && (reverse != CROSS_ENDIAN) ? SwapBytes(v,size) : v; };
            // Read full 32 bits from the memory/device,
            // unless we're going to replace it entirely.
            MemDataRef<size> data;
            if(rw != intent_store || size != 4)
            {
                NOfprintf(stderr,"IOread(%08X) = ", (unsigned)phys_addr);
                data.d.value = IOmap[phys_addr/0x10000].first(phys_addr & ~3);
                NOfprintf(stderr, "%08X\n", (unsigned)data.d.value);
            }
            // Create a reference to the relevant data
            auto& r = data.reffun( (phys_addr & (4-size)) ^ (reverse ? (4-size) : 0) );
            if(rw==intent_store)
            {
                // Write to the relevant data, and commit
                // the entire 32-bit word the memory/device. Byteswap if needed.
                r = S(value);
                NOfprintf(stderr,"IOwrite(%08X,%08X)\n",(unsigned)phys_addr,(unsigned)data.d.value);
                IOmap[phys_addr/0x10000].second(phys_addr & ~3, data.d.value);
            }
            else
                value = S(r); // Read the relevant data. Byteswap if needed.
        }
        catch(const std::bad_function_call& c)
        {
            fprintf(stderr, "MemAccess(virt=%08X,phys=%08X,size=%u,reverse=%s,mode=%s,UM=%d,VM=%d> failed because of bad function call!\n",
                virt_addr, phys_addr, size, reverse?"true":"false", rw==intent_store?"write":"read", regs.UM, regs.VM);
        }
        return 0;
    }

    class Tracer
    {
        std::list<std::pair<u32,unsigned> > stack;
    public:
        unsigned in_heavy;

        static unsigned heavy_locations_find(u32 addr)
        {
            static std::map<u32, unsigned> heavy_locations;
            if(heavy_locations.empty())
            {
                unsigned bitmask = 1;
                for(const auto i: stab)
                    if(i.second == "T memset"
                    || i.second == "T memcpy"
                    || i.second == "T calibrate_delay"
                    || i.second == "T calibrate_delay"
                    || i.second == "T printk"
                    || i.second == "T free_all_bootmem"
                    || i.second == "T memmove"
                    || i.second == "T strcpy"
                    || i.second == "T strcmp"
                    || i.second == "T sort"
                    || i.second == "T mapin_ram")
                    {
                        heavy_locations.insert( {(u32) i.first, bitmask} );
                        bitmask <<= 1;
                    }
            }
            auto i = heavy_locations.find(addr);
            if(i == heavy_locations.end()) return 0;
            return i->second;
        }

        void Push(u32 target, unsigned mode)
        {
            stack.push_back( {target,mode} );
            in_heavy |= heavy_locations_find(target);
        }
        void Pop(unsigned mode)
        {
            if(!stack.empty())
            {
                in_heavy &= ~ heavy_locations_find( stack.back().first );
                stack.pop_back();
            }
        }
        std::string Info() const
        {
            std::string result;
            for(auto i = stack.rbegin(); i != stack.rend(); ++i)
            {
                const auto& p = *i;
                unsigned addr = p.first;
                auto i = stab.find(addr);
                if(i == stab.end() && addr < 0xC0000000u)
                {
                    addr = addr - 0xC0000000 + 0x10000000;
                    i = stab.find(addr);
                }
                char Buf[64];
                sprintf(Buf, "%08X", addr);
                if(!result.empty()) result += "^";
                if(i != stab.end())
                {
                    const std::string& s = i->second;
                    if(s[1] == ' ')
                        result += s.substr(2);
                    else
                        result += s;
                }
                else
                    result += Buf;
                if(p.second == 1) result += "[I]";
                if(p.second == 2) result += "[B]";
                if(p.second == 3) result += "[B]";
                if(p.second == 4) result += "[E]";
            }
            return result;
        }
        size_t Size() const { return stack.size(); }
        bool is_in_heavy() const
        {
            return in_heavy != 0;
        }
        std::string GetString(u32 regval) const
        {
            std::string result;
            bool access_problem = false;
            for(unsigned c=0; regval >= 0x10000000 && c<300;++c)
            {
                extern MicroBlazeCPU cpu;
                u32 b=0;
                unsigned addr = regval + c;
                if(cpu.MemAccess<intent_load,1> (addr, b))
                {
                    access_problem = true;
                    break;
                }
                if(!b) break;
                if(b == '\n') result += "\\n";
                else if(b == '\r') result += "\\r";
                else if(b == '\b') result += "\\b";
                else if(b == '\t') result += "\\t";
                else if(b == '\0') result += "\\0";
                else result += (char)b;
            }
            result.insert(0, 1, '"');
            if(access_problem) result.insert(0, 1, '*');
            char Buf[64];
            sprintf(Buf, "\"(%08X)", regval);
            result += Buf;
            return result;
        }
    } Trace;

    void Debug_Syscall()
    {
        auto i = syscalls.find(active_syscall.regs[12]);

        fprintf(stderr, "[PID=%02X]: %s(", active_syscall.pid, i->second.name.c_str());
        int regno = 5;
        const char* sep = "";
        for(auto j: i->second.params)
        {
            fprintf(stderr, "%s", sep); sep = ", ";
            if(!j.name.empty())
                fprintf(stderr, "%s:", j.name.c_str());
            switch(j.type)
            {
                case Syscall::param::m_int:
                    if(j.name == "flags" || j.name == "mode")
                        fprintf(stderr, "0%o", active_syscall.regs[regno]);
                    else
                        fprintf(stderr, "0x%X", active_syscall.regs[regno]);
                    regno += 1;
                    break;
                case Syscall::param::m_long:
                    fprintf(stderr, "0x%llX",
                        (long long)((u64)active_syscall.regs[regno]
                                + (((u64)active_syscall.regs[regno+1]) << (u64)32)));
                    regno += 2;
                    break;
                case Syscall::param::m_charp:
                    fprintf(stderr, "%s", Trace.GetString(active_syscall.regs[regno++]).c_str());
                    break;
                case Syscall::param::m_charpp:
                {
                    unsigned ptrptr = active_syscall.regs[regno++], ptrptr1 = ptrptr;
                    fprintf(stderr, "{");
                    const char* sep = "";
                    for(;;)
                    {
                        fprintf(stderr, "%s", sep); sep = ",";
                        u32 ptr;
                        if(MemAccess<intent_load,4> (ptrptr, ptr, false))
                        {
                            fprintf(stderr, "*");
                            break;
                        }
                        if(!ptr) break;
                        fprintf(stderr, "%s", Trace.GetString(ptr).c_str());
                        ptrptr += 4;
                    }
                    fprintf(stderr, "}(%08X)", ptrptr1);
                    break;
                }
                default:
                    fprintf(stderr, "0x%08X", active_syscall.regs[regno++]);
                    break;
            }
        }
        fprintf(stderr, ")");
    }

    void OpTrace()
    {
        /*if(Trace.is_in_heavy())
            OPCODE_DEBUG_ENABLED = false;
        else
            OPCODE_DEBUG_ENABLED = true;*/

        auto i = stab.find( regs.VM ? regs.PC
                                    :(regs.PC - 0x10000000 + 0xC0000000) );
   #if OPCODE_DEBUG || 1
        if(i != stab.end())
        {
            fprintf(stderr, "\33[1;32m%s: %*s%s (%s)\33[0;37m...\n",
                regs.VM ? "MM" : "RAW",
                (unsigned)Trace.Size(),"",
                i->second.c_str(),
                ""/*Trace.Info().c_str()*/);
            /*if(i->second == "T __vm_enough_memory")
                OPCODE_DEBUG_ENABLED = true;*/
            if(i->second == "T pidmap_init")
            {
                // A little anointing hack. It's a bit easier
                // to debug when not constantly interrupted
                // by timer interrupts.
                TIMER_ENABLED = false;
            }
            if(i->second == "T emit_log_char")
            {
                Console.Putc(regs.r[5]);
            }
        }
        if(i != stab.end() && (i->second == "T panic"
                            || i->second == "T die"
                            || i->second == "T vprintk"
                            || i->second == "t tty_write"
                            || i->second == "T kupo"
                            || i->second == "T do_execve"
                            || i->second == "T early_printk"
                            || i->second == "t early_printk_uartlite_write"
                            || i->second == "t early_printk_uart16550_write"))
        {
            // R7 = number of characters
            // R6 = pointer to data, R4 = pointer to early_serial_uartlite_console
            fprintf(stderr, "Reached '%s' with R5=%s, R6=%s, R7=%s\n",
                i->second.c_str(), Trace.GetString(regs.r[5]).c_str(),
                                   Trace.GetString(regs.r[6]).c_str(),
                                   Trace.GetString(regs.r[7]).c_str());
            fflush(stderr);
        }
    #endif
        if(i != stab.end() && i->second == "t no_blink")
        {
            fprintf(stderr, "KERNEL CRASHED, REACHED no_blink\n");
            exit(-1);
        }
    }

    // This struct records the desired action to happen after instruction execution.
    struct execmode
    {
        unsigned mode:2, type:30, target:32;
        // Modes:
        //   0 = Normally proceed to the next instruction (other fields are ignored)
        //   1 = Jump, right now. (type=type of jump, target=where to jump)
        //   2 = Jump, but first execute one more instruction (delay slot)
        //   3 = Generate an exception, right now
        //       (type   = type of exception,
        //        target = memory address causing the exception)
    };

    // Decode and execute one instruction.
    execmode ExecuteOne(u32 i, u32 arg2)
    {
#if VERIFY_RAM
        if(regs.r[0] != 0) // Sanity check
            abort();
#endif
        // Decode the operation type and operands from the opcode
        unsigned i_op = (i>>26)&0x3F;  // Operation
        unsigned i_rd = (i>>21)&0x1F;  // Target register index
        u32 RA = regs.r[(i>>16)&0x1F]; // Source register value
        bool imm = i_op & 0x08;        // Category of instruction
        if(!imm) arg2 = regs.r[(i>>11)&0x1F]; // Second operand

        // Exit macros.
        // The RD family exits write a value into RD, unless happens to be 0.
        // r[0] is never overwritten; it stays as 0.
        #define Return return execmode{0,0,0}
        // Exit and store v to RD. Execute optional code x.
        #define ReturnRDx(v,x) do { if(i_rd) regs.r[i_rd] = (v); \
                                    if(OPCODE_DEBUG_ENABLED) \
                                    { \
                                       unsigned r = regs.r[i_rd]; \
                                       fprintf(stderr, "-> R%u=%08X '%c'\n", i_rd, r, \
                                        r>=0x20 && r<=0x7E ? r : '?'); \
                                    } x Return; } while(0)
        // Exit and store v to RD.
        #define ReturnRD(v) ReturnRDx(v,)
        // Exit and store v to RD, and set carry to c.
        #define ReturnRDc(v,c) do { bool nc = (c); ReturnRDx(v, regs.C = regs.CC = nc;); } while(0)
        // Exit and prepare to branch/jump to another memory location
        #define ReturnJmp(target,delayed,type) return execmode{(delayed)?2:1, type, target}
        // Exit and raise an exception, concerning the given address.
        #define ReturnEx(extype, addr) return execmode{3,extype,addr}
        // Exit and raise a MMU exception, setting the exception info appropriately.
        #define ReturnMMUex(extype, ear, write) \
            do { regs.Rx = i_rd; regs.S = write; regs.DIZ = error & 4; \
                 ReturnEx(extype & ~4, ear); } while(0)

        // Determine which kind of operation it was:
        if(i == 0x00000000) ReturnEx(2,0); // Specially trap this illegal opcode
        #define tr(op) (((op) & 0x07) | (((op) & 0x30) >> 1))
        switch( tr(i_op) )
        {
            // Add and subtract operations: (ADD|RSUB)(I)(K)(C); comparisons: CMP(U)
            case tr(0x06): if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
                           ReturnRD(  arg2+RA+regs.C                          ); // ADD(I)KC
            case tr(0x07): if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
                           ReturnRD(  arg2-RA-regs.C                          ); // RSUB(I)KC
            case tr(0x02): if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
                           ReturnRDc( arg2+RA+regs.C, arg2+RA+regs.C   < arg2 ); // ADD(I)C
            case tr(0x03): if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
                           ReturnRDc( arg2-RA-regs.C, arg2-RA-regs.C   > arg2 ); // RSUB(I)C
            case tr(0x00): if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
                           ReturnRDc( arg2+RA,        arg2+RA          < arg2 ); // ADD(I)
            case tr(0x01): if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
                           ReturnRDc( arg2-RA,        arg2-RA          > arg2 ); // RSUB(I)
            case tr(0x04): if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
                           ReturnRD(  arg2+RA                                 ); // ADD(I)K
            case tr(0x05): switch(imm?0:(i&0x7FF))
            { case 0:      ReturnRD(  arg2-RA                                 ); // RSUB(I)K
              case 1: ReturnRD((0x7FFFFFFF&(arg2-RA)) | (((s32)arg2 < (s32)RA)<<31) ); // CMP
              case 3: ReturnRD((0x7FFFFFFF&(arg2-RA)) | (((u32)arg2 < (u32)RA)<<31) ); // CMPU
              default:     if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
            } Return;/*Silently ignore bad instruction*/

            // Bit manipulation operations: (OR|AND|XOR|ANDN)(I)
            case tr(0x20): if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
                           ReturnRD( RA |  arg2 ); // OR,  ORI
            case tr(0x21): if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
                           ReturnRD( RA &  arg2 ); // AND, ANDI
            case tr(0x22): if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
                           ReturnRD( RA ^  arg2 ); // XOR, XORI
            case tr(0x23): if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
                           ReturnRD( RA &~ arg2 ); // ANDN,ANDNI
#if C_USE_BARREL || 1
            // Barrel shift operations: BS(RL|RA|LL)(I)
            case tr(0x11):
                if(i&0x400) ReturnRD( RA      << (arg2 & 31) ); // BSLL(I)
                if(i&0x200) ReturnRD( s32(RA) >> (arg2 & 31) ); // BSRA(I)
                else        ReturnRD( u32(RA) >> (arg2 & 31) ); // BSRL(I)
#endif
#if C_USE_HW_MUL || 1
            // Multiplications: MUL(I|H|HU|HSU)
            case tr(0x10): switch(imm ? 0 : (i&0x7FF))
            { case 0: ReturnRD( (s32)RA * (s32)arg2 );                   // MUL,MULI
              case 1: ReturnRD( ((s64)(s32)RA * (s64)(s32)arg2) >> 32 ); // MULH
              case 2: ReturnRD( ((s64)(s32)RA * (s64)(u32)arg2) >> 32 ); // MULHSU
              case 3: ReturnRD( ((u64)(u32)RA * (u64)(u32)arg2) >> 32 ); // MULHU
              default:     if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
            } Return;/*Silently ignore bad instruction*/
#endif
            // One-bit copy / shift-right operators: SEXT(8|16), SR(A|C|L); cache control: W(I|D)C
            case tr(0x24): switch(i & 0x7FF)
            { case 0x60: ReturnRD( (s32)( s8)RA ); // SEXT8  (sign-extend 8 bits to 32 bits)
              case 0x61: ReturnRD( (s32)(s16)RA ); // SEXT16 (sign-extend 16 bits to 32 bits)
              case 0x01: ReturnRDc(            (s32)RA >> 1,  RA&1); // SRA (sign-extend)
              case 0x21: ReturnRDc((regs.CC<<31) | (RA >> 1), RA&1); // SRC (carry-extend)
              case 0x41: ReturnRDc(                 RA >> 1,  RA&1); // SRL (zero-extend)
              case 0x68: if(regs.UM) ReturnEx(7,0); Return; // WIC
              case 0x64: if(regs.UM) ReturnEx(7,0); Return; // WDC
              case 0x66: if(regs.UM) ReturnEx(7,0); Return; // WDC.clear
              case 0x74: if(regs.UM) ReturnEx(7,0); Return; // WDC.flush
              case 0x76: if(regs.UM) ReturnEx(7,0); Return; // WDC.flush.clear
              default:     if(!imm && (i & 0x7FF)) fprintf(stderr, "ERROR: TRIED subfunction %X on i_op=%02X\n", i&0x7FF,i_op);
            } Return;/*Silently ignore bad instruction*/
            // Conditional branches with delay slots or not: (BEQ|BNE|BLT|BLE|BGT|BGE)(I)(D)
            case tr(0x27): switch(i_rd & 15)
            { case 0: if((s32)RA == 0) ReturnJmp( regs.PC-4 + arg2, i_rd & 16, 0 ); Return; //EQ
              case 1: if((s32)RA != 0) ReturnJmp( regs.PC-4 + arg2, i_rd & 16, 0 ); Return; //NE
              case 2: if((s32)RA  < 0) ReturnJmp( regs.PC-4 + arg2, i_rd & 16, 0 ); Return; //LT
              case 3: if((s32)RA <= 0) ReturnJmp( regs.PC-4 + arg2, i_rd & 16, 0 ); Return; //LE
              case 4: if((s32)RA  > 0) ReturnJmp( regs.PC-4 + arg2, i_rd & 16, 0 ); Return; //GT
              case 5: if((s32)RA >= 0) ReturnJmp( regs.PC-4 + arg2, i_rd & 16, 0 ); Return; //GE
            } Return;/*Silently ignore bad instruction*/
            // Unconditional branches / system calls, with delay slots or not: BR(A)(L)(I)(D), BRK(I), MBAR
            case tr(0x26): switch( (i >> 18) & 7) // This is hairy, because there are so many variants in this.
            {
                /*
                101110 _imm_ 00010 00000 00000000100 MBAR = immediate relative jump of +4
                100110 __rD_ 01100 __rB_ 00000000000 BRK
                101110 __rD_ 01100 iiiii iiiiiiiiiii BRKI
                100110 __rD_ daL00 __rB_ 00000000000 BR(A)(L)(D)
                101110 __rD_ daL00 iiiii iiiiiiiiiii BR(A)(L)(I)(D)
                                  Valid combinations:   BR,  BRA, BRAD, BRD, BRLD, BRALD
                                  Invalid combinations: BRL, BRAL
                */
                case 4: // D=1, A=0, L=0 (BR(I)D)
                    if(i_rd) fprintf(stderr, "BRD: Spurious, RD=%u??\n",i_rd);
                    ReturnJmp(regs.PC-4+arg2, true, 0);
                case 0: // D=0, A=0, L=0 (BR(I) or MBAR).
                    if(i_rd) fprintf(stderr, "BR/MBAR: Spurious, RD=%u??\n",i_rd);
                    ReturnJmp(regs.PC-4+arg2, false, 0);
                case 5: // D=1, A=0, L=1 (BRL(I)D)
                    if(i_rd) regs.r[i_rd] = regs.PC - 4;
                    if(i_rd) Trace.Push( regs.PC-4+arg2, 0 );
                    if(!i_rd) fprintf(stderr, "BRLD: Spurious, RD=0??\n");
                    ReturnJmp(regs.PC-4+arg2, true, 0);
                case 1: // D=0, A=0, L=1 (BRL(I) = invalid)
                    break;
                case 6: // D=1, A=1, L=0 (BRA(I)D)
                    if(i_rd) fprintf(stderr, "BRAD: Spurious, RD=%u??\n",i_rd);
                    ReturnJmp(arg2, true, 0);
                case 2: // D=0, A=1, L=0 (BRA(I))
                    if(i_rd) fprintf(stderr, "BRA: Spurious, RD=%u??\n",i_rd);
                    ReturnJmp(arg2, false, 0);
                case 7: // D=1, A=1, L=1 (BRAL(I)D; escalate if 0x08)
                   {bool escalate = imm && arg2 == v_usr_ex;
                    if(i_rd) regs.r[i_rd] = regs.PC - 4;
                    if(i_rd) Trace.Push( arg2, escalate ? 3 : 0 );
                    if(!i_rd) fprintf(stderr, "BRALD: Spurious, RD=0??\n");
                    ReturnJmp(arg2, true, escalate ? 3 : 0 );}
                case 3: // D=0, A=1, L=1 (BRK(I) = special; escalate if 0x08 or 0x18, otherwise deny UM)
                   {bool escalate = imm && (arg2 == v_usr_ex || arg2 == v_brk);
                    if(regs.UM && !escalate) ReturnEx(7,0);
                    if(i_rd) regs.r[i_rd] = regs.PC - 4;
                    if(i_rd) Trace.Push( arg2, escalate ? 3 : 0 );
                    if(!i_rd) fprintf(stderr, "BRK: Spurious, RD=0??\n");
                    regs.BIP = true;
                    fprintf(stderr, "Generating BREAK to address: %X, R12=%d, R5=%08X,R6=%08X,R7=%08X,R8=%08X,R9=%08X,R10=%08X\n",
                        arg2, regs.r[12],regs.r[5],regs.r[6],regs.r[7],regs.r[8],regs.r[9],regs.r[10]);
                    auto i = syscalls.find(regs.r[12]);
                    if(i != syscalls.end())
                    {
                        active_syscall.return_pc = regs.PC;
                        active_syscall.pid       = regs.PID;
                        for(unsigned a=0; a<32; ++a)
                            active_syscall.regs[a] = regs.r[a];

                        fprintf(stderr, "SYSCALL");
                        Debug_Syscall();
                        fprintf(stderr, "\n");
                    }
                    ReturnJmp(arg2, false, escalate ? 3 : 0 );}
            }Return;/*Silently ignore bad instruction*/
            // Subroutine/exception returns: RT(S|I|B|E)D; Special register manipulation: M(F|T)S, MSR(CLR|SET)
            case tr(0x25):
                if(imm) // Note: MicroBlaze's return opcodes always include a delay slot.
                {
                    // B=2, I=1, E=4, S=0
                    if((i_rd&7)==0) Trace.Pop(0);
                    if((i_rd&7) && regs.UM) ReturnEx(7,0); // privileged return
                    ReturnJmp(RA+arg2, true, i_rd&7); // RTSD/RTID/RTBD/RTED
                }
               {u32& op2 = i_rd ? regs.r[i_rd] : arg2;
                switch( (i >> 14) & 7)
                {
                    // MFS:    100101 __rD_ 00000 10 rsrsrsrsrsrsrs    010, 110
                    // MTS:    100101 00000 __rA_ 11 rsrsrsrsrsrsrs    011, 111
                    // MSRSET: 100101 __rD_ 10000 0 iiiiiiiiiiiiiii    000, 001
                    // MSRCLR: 100101 __rD_ 10001 0 iiiiiiiiiiiiiii    100, 101
#if C_USE_MSR_INSTR || 0
                    case 0: case 1: // MSRSET
                        if(regs.UM && ((i&0x7FFF) != 4)) ReturnEx(7,0);
                        op2 = regs.MSR; regs.MSR |= (i&0x7FFF);
                        break;
                    case 4: case 5: // MSRCLR:
                        if(regs.UM && ((i&0x7FFF) != 4)) ReturnEx(7,0);
                        op2 = regs.MSR; regs.MSR &= ~(i&0x7FFF);
                        break;
#endif
                    case 2: case 6: 
                        regs.SPR( i&0x3FFF, false, op2); // MFS
                        break;
                    case 3: case 7: // MTS
                        if(regs.UM) ReturnEx(7,0);
                        regs.SPR( i&0x3FFF, true, RA ); // MTS
                }
                // Update the carry-copy bit
                regs.CC = regs.C;
                regs.PVR = true; // Ensure this flag is never turned off
                ReturnRD(regs.r[i_rd]);}

            // Memory access operations:
            #define HandleLoadStoreOpcode(opcode, rw, size) \
                case tr(opcode): \
                { \
                    if(OPCODE_DEBUG_ENABLED) \
                        fprintf(stderr, "-> address = %08X\n", RA+arg2); \
                    if(auto error = MemAccess<rw,size>(RA+arg2, regs.r[i_rd], !imm && (i&0x200))) \
                        ReturnMMUex(error, RA+arg2, rw==intent_store); \
                    if(rw == intent_load && !i_rd) regs.r[i_rd] = 0; \
                    if(!imm && (i&0x400)) regs.C = regs.CC = false; /* exclusive access, not implemented */ \
                    ReturnRD(regs.r[i_rd]); \
                }
            HandleLoadStoreOpcode(0x30, intent_load,1) // LBU(I)(R)
            HandleLoadStoreOpcode(0x31, intent_load,2) // LHU(I)(R)
            HandleLoadStoreOpcode(0x32, intent_load,4) // LW(I)(R)(X)
            HandleLoadStoreOpcode(0x34, intent_store,1) // SB(I)(R)
            HandleLoadStoreOpcode(0x35, intent_store,2) // SH(I)(R)
            HandleLoadStoreOpcode(0x36, intent_store,4) // SW(I)(R)(X)
            // C_USE_PCMP_INSTR not implemented (PCMPNE,PCMPEQ,PCMPBF,CLZ) because CLZ is too complex, PCMPBF is also quite complex
            // C_USE_DIV        not implemented (IDIV,IDIVU) because checking for the corner cases is too complex
            // C_USE_FPU=1      not implemented (FADD,FRSUB,FMUL,FDIV,FCMP) because lots of code (comparisons, corner cases)
            // C_USE_FPU=2      not implemented (FLT,FINT,FSQRT) because C_USE_FPU=1 not implemented
            // C_USE_STACK_PROTECTION not implemented because not utilized by Linux
            // C_FSL_LINKS      not implemented because huge function set, and not utilized by Linux
        }
        #undef tr
        // Anything else: Illegal instruction exception
        fprintf(stderr, "ERROR: Illegal instruction %08X\n", i);
        ReturnEx(2,0);
    }

    void DoReset(u32 vector = v_reset, u32 cmdline_pointer = 0)
    {
        // Clear all registers. (Note: Real MicroBlaze does not clear R1..R31.)
        regs = {};
        regs.PVR = true;
        regs.EAR = regs.ESR = 0;
        regs.PID = regs.ZPR = regs.TLBX = 0;
        regs.SLR = 0; regs.SHR = 0xFFFFFFFFu;
        // Set up bit fields indicating the list of processor features
        #define Fln(index, n,bit, why) \
            regs.PVRs[index] |= ((n) << (31-(bit)));
        #define Fl(index, bit, why) Fln(index, 1,bit,why)
        Fl(0, 0, "C_PVR")
        Fl(0, 5, "C_USE_MMU > 0, exceptions enabled")
        Fl(0, 8, "C_USE_MMU > 0")
        if(LITTLE_ENDIAN_SLAVE) Fl(0, 10, "little endian")
        Fln(0, 0x15, 23,  "Release version code 8.20.a")
    #if C_USE_MSR_INSTR || 0
        Fl(2, 14, "C_USE_MSR_INSTR")
    #endif
    #if C_USE_BARREL || 1
        Fl(0, 1, "C_USE_BARREL")
        Fl(2, 17, "C_USE_BARREL")
    #endif
    #if C_USE_HW_MUL || 1
        Fl(0, 3, "C_USE_HW_MUL > 0")
        Fl(2, 19, "C_USE_HW_MUL > 0")
        Fl(2, 21, "C_USE_HW_MUL = 2")
    #endif
        Fl(2, 25, "C_OPCODE_0x0_ILLEGAL")
    #if C_UNALIGNED_EXCEPTIONS || 1
        Fl(2, 26, "C_UNALIGNED_EXCEPTIONS");
    #endif
        Fl(2, 27, "C_ILL_OPCODE_EXCEPTION")
        Fln(11, 3, 1, "C_USE_MMU = 3")
        Fln(11, 3, 9, "C_MMU_TLB_ACCESS = 3")
        Fln(11, 16, 14, "C_MMU_ZONES = 16")
        Fl(11, 15, "C_MMU_PRIVILEGED_INSTR")
        Fln(11, regs.MSR, 31, "C_RESET_MSR")
        regs.PC = vector;
        regs.r[5] = cmdline_pointer;
    }

    bool InterruptsEnabled() const
    {
        return regs.IE && !regs.EIP && !regs.BIP;
    }
    void DoInterrupt(u32 vector = v_intr)
    {
        regs.IE = false;
        regs.PushUVM();
        regs.r[14] = regs.PC;
        regs.PC = vector;
        NOfprintf(stderr, "Interrupt...\n");
        Trace.Push(vector, 1);
    }

    void Singlestep()
    {
        execmode next;

        auto NextOpcode = [&] (u32& value) -> bool
        {
            unsigned error = MemAccess<intent_execute,4>(regs.PC, value, !CROSS_ENDIAN);
            if(error) { next = {3,error&~4,regs.PC}; return false; } // Exception

    #if OPCODE_DEBUG
            if(OPCODE_DEBUG_ENABLED)
            {
                /***/ // DEBUG
                static disassemble_info info; static bool i=false;
                if(!i)
                {
                    i = true;
                    info.read_memory_func =
                        [](bfd_vma,bfd_byte* myaddr,unsigned l,struct disassemble_info* info) -> int
                        {
                            memcpy(myaddr, &info->buffer_vma, l);
                            return 0;
                        };
                    info.endian = BFD_ENDIAN_BIG;
                    info.fprintf_func = (fprintf_ftype) fprintf;
                    info.stream = stderr;
                    info.print_address_func = paddr;
                    info.symbol_at_address_func = sfunc;
                }
                /**/ // DEBUG
                fprintf(stderr, "\33[0;33m");
                paddr(regs.PC, 0);
                fprintf(stderr, "[PID=%02X]\t", regs.PID);
                info.buffer_vma = (CROSS_ENDIAN ? SwapBytes(value,4) : value);
                fprintf(stderr, "%08X\t", value);
                print_insn_microblaze(regs.PC, &info);
                fprintf(stderr, "\33[0m\n");
            }
    #endif
            regs.PC += 4;
            return true;
        };

        OPCODE_DEBUG_ENABLED = regs.UM && DEBUG_FORCED;

        u32 op_begin = regs.PC;
        if(!regs.BIP) OpTrace();

        u32 opcode;
        if(!NextOpcode(opcode)) goto GotException;

       {s32 immval = (s16)opcode;
        // If the opcode was IMM, read the next one immediately.
        if( (opcode >> 16) == 0xB000 )
        {
            if(!NextOpcode(opcode)) goto GotException;
            immval = (immval << 16) | (opcode & 0xFFFF); // Extend the literal parameter
        }
        next = ExecuteOne(opcode, immval);}
        switch(next.mode)
        {
            case 0: default: break;
            case 2: // Delayed jump
            {
                u32 btr = next.target;
                if(!NextOpcode(opcode)) { next.mode = 2; regs.BTR = btr; goto GotException; }

                // Execute the in-delay-slot instruction.
                // Note: IMM and branches cannot exist in delay slot.
                execmode next2 = ExecuteOne(opcode, (s16)opcode);
                if(next2.mode == 3) // Did an exception happen in a delay slot?
                    { regs.BTR    = btr;
                      next.target = next2.target; // Copy exception address
                      next.type   = next2.type;   // Copy exception type
                      goto GotException; }
            } // passthru
            case 1: // Jump
#if OPCODE_DEBUG
                if(OPCODE_DEBUG_ENABLED)
                {
                    fprintf(stderr, "[PC=%08X]Dealing with jump to %08X, type=%u, delay=%s\n",
                        regs.PC, next.target, next.type, next.mode==2?"was":"no");
                }
#endif
                switch(next.type) // Handle possible special types of jumps
                {
                    // type 1 = Return From Interrupt (set IE   ; copy UMS,VMS to UM,VM)
                    case 1: Trace.Pop(1); regs.IE = true; regs.PopUVM(); break;
                    // type 2 = Return From Break (clear BIP; copy UMS,VMS to UM,VM)
                    case 2: Trace.Pop(2);
                            if(next.target >= 0x00001000 && next.target < 0xC0000000u)
                            fprintf(stderr, "Exiting from %s, to %08X, PC=%08X, R3=%08X\n",
                                regs.BIP ? "BREAK" : "something", next.target, regs.PC, regs.r[3]);
                            regs.BIP = false; regs.PopUVM();
                            if(next.target == active_syscall.return_pc)
                            {
                                fprintf(stderr, "SYSCALL");
                                Debug_Syscall();
                                fprintf(stderr, " RETURN VALUE: %d\n", regs.r[3]);
                            }
                            break;
                    // type 3 = Break or User Exception (copy UM,VM to UMS,VMS)
                    case 3: regs.PushUVM(); break;
                    // type 4 = Return From Exception (set EE, clear EIP, zero ESR; copy UMS,VMS to UM,VM)
                    case 4: Trace.Pop(4); regs.EE = true; regs.EIP = false; regs.ESR = 0; regs.PopUVM(); break;
                }
                regs.PC = next.target;
                break;
            case 3: GotException:; // Exception
                unsigned extype = next.type, ear = next.target;

                fprintf(stderr, "GENERATING EXCEPTION type %u (PID=%02X, EE=%d, EIP=%d, BIP=%d, IE=%d, UM=%d, VM=%d, EAR=%08X, PC=%08X..%08X)\n",
                    extype, regs.PID, regs.EE, regs.EIP, regs.BIP, regs.IE, regs.UM, regs.VM, ear, op_begin, regs.PC);
                if(regs.EE || extype >= 16)
                {
                    // If EE is unset, ignore exception unless caused by MMU
                    // For MMU exceptions, the return address is the beginning
                    // of the opcode that caused the fault. For others, it is
                    // the next opcode.
                    regs.EIP = true;
                    regs.EAR = ear;
                    regs.EE  = false;
                    regs.EC  = extype;
                    regs.DS  = next.mode == 2; // was it a delay slot exception?
                    if(regs.DS)
                        fprintf(stderr, "NOTE: DELAY SLOT EXCEPTION\n");
                    regs.PushUVM();
                    regs.r[17] = (extype < 16 ? regs.PC : op_begin);
                    regs.PC = v_hw_ex;
                    Trace.Push(v_hw_ex, 4);
                }
        }
    }
};

// The DTS file embedded with this Linux kernel
// has the following presets configured in it:
MicroBlazeCPU  cpu;      // Configured at 100 MHz
SDRAM<0x01000000> RAM;   // IO: 10000000 , 16 MB of SD-RAM
SDRAM<0x10000> LowRAM;   // IO: 0x000000 , 64 kB of RAM (needed for exception tables)
XilinxINTC        intc;  // IO: 81800000
XilinxTimer<2>    timer; // IO: 83c00000 IRQ 3
XilinxUARTlite    uart;  // IO: 84000000 IRQ 8

static u32 S(u32 v)
{
    return CROSS_ENDIAN ? SwapBytes(v, 4) : v;
}

int main()
{
/***/ // DEBUG
   {FILE* fp = fopen("linux-3.0.4/System.map", "rt");
    for(;;)
    {
        char Buf[1024];
        if(!fgets(Buf, sizeof(Buf), fp)) break;
        unsigned long addr;
        char mode;
        sscanf(Buf, "%lX %c %s", &addr, &mode, Buf+2);
        Buf[0] = mode;
        Buf[1] = ' ';
        stab[addr] = Buf;
    }
    fclose(fp);}

   {std::map<std::string, unsigned> syscall_numbers;
    FILE* fp = popen("f=linux-3.0.4/arch/microblaze/include/asm/unistd.h; "
                     "(cat $f;sed 's@/\\*.*\\*/@@' < $f|grep __NR_|perl -pe 's/.*__NR_([a-z0-9_]*).*/__NR_$1 $1/')|gcc -E -",
                     "r");
    for(;;)
    {
        char Buf[1024];
        if(!fgets(Buf, sizeof(Buf), fp)) break;
        unsigned num;
        sscanf(Buf, "%u %s", &num, Buf);
        syscall_numbers.insert( {Buf,num} );
    }
    pclose(fp);
    for(auto n: {"linux","asm-generic"})
   {
    std::string cmd = "grep -v '#inc' linux-3.0.4/include/";
    cmd += n;
    cmd += "/syscalls.h |gcc -DBITS_PER_LONG=32 -E - |grep -v '#'|tr '\\012' ' '|tr ';' '\\012'"
           "|grep asmlinkage|sed 's/^ *asmlinkage *//;s/__user//g;s/^long sys_/:/'|tr '(),' '\\012\\012\\012'|sed 's/^ *//;s/  */ /g'";
    fp = popen(cmd.c_str(), "r");
    for(unsigned current=~0;;)
    {
        char Buf[1024];
        if(!fgets(Buf, sizeof(Buf), fp)) break;
        for(char*s=Buf; *s; ++s) if(*s=='\r' || *s=='\n') { *s=0; break; }
        if(Buf[0] == ':')
        {
            current = syscall_numbers[Buf+1];
            syscalls[current].no = current;
            syscalls[current].name = Buf+1;
            continue;
        }
        Syscall& sys = syscalls[current];
        sys.no   = current;
        if(std::string(Buf) == "void") continue;
        std::string name;
        char* s = Buf, *p = std::strrchr(Buf, '\0');
        while(p > s && p[-1] != ' ' && p[-1] != '*') name.insert(0, 1, *--p);
        while(p > s && p[-1] == ' ') --p;
        *p = 0;
        std::string type = Buf;
        if(p == s) { type = name; name = ""; }
        if(type.empty()) continue;

        if(type == "int" || type == "unsigned int" || type == "short" || type == "unsigned short"
        || type == "pid_t" || type == "gid_t" || type == "uid_t" || type == "size_t"
        || type == "ssize_t" || type == "off_t" || type == "long" || type == "unsigned long")
            sys.params.push_back( { name, Syscall::param::m_int } );
        else if(type == "off64_t" || type == "loff_t" || type == "u64")
            sys.params.push_back( { name, Syscall::param::m_long } );
        else if(type == "char *" || type == "const char *")
            sys.params.push_back( { name, Syscall::param::m_charp } );
        else if(type == "const char *const *")
            sys.params.push_back( { name, Syscall::param::m_charpp } );
        else
            sys.params.push_back( { name, Syscall::param::m_ptr } );
    }}
    for(auto i: syscalls)
    {
        fprintf(stderr, "syscall %u: %s", i.second.no, i.second.name.c_str());
        for(auto j: i.second.params)
            fprintf(stderr, "| %s:%d", j.name.c_str(), (int) j.type);
        fprintf(stderr, "\n");
    }
    pclose(fp);
   }

/***/
    // Set up the memory mapping
    cpu.IOmap[0x0000] = { [](u32 i)       { return LowRAM.Read(i/4); } ,
                          [](u32 i,u32 v) { LowRAM.Write(i/4,v); } };
    for(unsigned a=0x1000; a<0x1200; ++a)
        cpu.IOmap[a] = { [](u32 i)       { return RAM.Read((i-0x10000000)/4); } ,
                         [](u32 i,u32 v) { RAM.Write((i-0x10000000)/4,v); } };

    cpu.IOmap[0x8180] = { [](u32 i)       { return S(intc.Read((i/4)%8)); } ,
                          [](u32 i,u32 v) { intc.Write((i/4)%8,S(v)); } };

    cpu.IOmap[0x83C0] = { [](u32 i)       { return S(timer.Read((i/16)&1, (i/4)%4)); } ,
                          [](u32 i,u32 v) { timer.Write((i/16)&1, (i/4)%4, S(v)); } };

    cpu.IOmap[0x8400] = { [](u32 i)       { return S(uart.Read((i/4)%4)); } ,
                          [](u32 i,u32 v) { uart.Write((i/4)%4,S(v)); } };


    // Load Linux kernel from U-Boot image
    struct { u32 a[3], image_size, load_offs, entry_offs, b[10]; } Header;
    FILE* fp = fopen("simpleImage.emu.ub", "rb");
    fread(&Header, 0x40, 1, fp);
    if(LITTLE_ENDIAN_HOST) Header.load_offs  = SwapBytes(Header.load_offs,  4);
    if(LITTLE_ENDIAN_HOST) Header.entry_offs = SwapBytes(Header.entry_offs, 4);
    if(LITTLE_ENDIAN_HOST) Header.image_size = SwapBytes(Header.image_size, 4);

    // Reset the CPU (and bring memory mapper to user mode)
    cpu.DoReset(Header.entry_offs, Header.load_offs + Header.image_size);

    // Read the kernel image to RAM
    for(auto offs = Header.load_offs; offs < Header.load_offs + Header.image_size; offs += 4)
    {
        u32 value;
        if(fread(&value, 1, 4, fp) <= 0) break;
        cpu.MemAccess<intent_store,4> (offs, value, CROSS_ENDIAN);
#if VERIFY_RAM
        u32 value2; /* DEBUG */
        cpu.MemAccess<intent_load,4> (offs, value2, CROSS_ENDIAN);
        if(value2 != value)
            fprintf(stderr,
                "RAM does not function properly! Stored %08X, got back %08X\n",
                value, value2);
#endif
    }
    fclose(fp);

    // Run the machine
    for(;;)
    {
        for(auto n = TIMER_ENABLED ? 1 : 10; n--; )
            cpu.Singlestep();

        uart.Input();
        timer.Tick();

        if(timer.CheckIRQ()) intc.TriggerIRQ(3);
        if(cpu.InterruptsEnabled())
        {
            if(uart.CheckIRQ()) intc.TriggerIRQ(8);
            if(intc.CheckIRQ()) cpu.DoInterrupt();
        }
    }
}
